from .build import build_rust
from .clean import clean_rust
from .extension import Binding, RustBin, RustExtension, Strip
from .version import version as __version__  # noqa: F401

__all__ = ("Binding", "RustBin", "RustExtension", "Strip", "build_rust", "clean_rust")



import subprocess
from typing import Any, Optional, Union, cast


class Env:
    """Allow using ``functools.lru_cache`` with an environment variable dictionary.

    Dictionaries are unhashable, but ``functools.lru_cache`` needs all parameters to
    be hashable, which we solve which a custom ``__hash__``."""

    env: Optional[dict[str, str]]

    def __init__(self, env: Optional[dict[str, str]]):
        self.env = env

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Env):
            return False
        return self.env == other.env

    def __hash__(self) -> int:
        if self.env is not None:
            return hash(tuple(sorted(self.env.items())))
        else:
            return hash(None)


def run_subprocess(
    *args: Any, env: Union[Env, dict[str, str], None], **kwargs: Any
) -> subprocess.CompletedProcess:
    """Wrapper around subprocess.run that requires a decision to pass env."""
    if isinstance(env, Env):
        env = env.env
    kwargs["env"] = env
    return subprocess.run(*args, **kwargs)  # noqa: TID251 # this is a wrapper to implement the rule


def check_subprocess_output(
    *args: Any, env: Union[Env, dict[str, str], None], **kwargs: Any
) -> str:
    """Wrapper around subprocess.run that requires a decision to pass env."""
    if isinstance(env, Env):
        env = env.env
    kwargs["env"] = env
    return cast(str, subprocess.check_output(*args, **kwargs))  # noqa: TID251 # this is a wrapper to implement the rule


def format_called_process_error(
    e: subprocess.CalledProcessError,
    *,
    include_stdout: bool = True,
    include_stderr: bool = True,
) -> str:
    """Helper to convert a CalledProcessError to an error message.

    If `include_stdout` or `include_stderr` are True (the default), the
    respective output stream will be added to the error message (if
    present in the exception).

    >>> format_called_process_error(subprocess.CalledProcessError(
    ...     777, ['ls', '-la'], None, None
    ... ))
    '`ls -la` failed with code 777'
    >>> format_called_process_error(subprocess.CalledProcessError(
    ...     1, ['cargo', 'foo bar'], 'message', None
    ... ))
    "`cargo 'foo bar'` failed with code 1\\n-- Output captured from stdout:\\nmessage"
    >>> format_called_process_error(subprocess.CalledProcessError(
    ...     1, ['cargo', 'foo bar'], 'message', None
    ... ), include_stdout=False)
    "`cargo 'foo bar'` failed with code 1"
    >>> format_called_process_error(subprocess.CalledProcessError(
    ...    -1, ['cargo'], 'stdout', 'stderr'
    ... ))
    '`cargo` failed with code -1\\n-- Output captured from stdout:\\nstdout\\n-- Output captured from stderr:\\nstderr'
    """
    command = " ".join(_quote_whitespace(arg) for arg in e.cmd)
    message = f"`{command}` failed with code {e.returncode}"
    if include_stdout and e.stdout is not None:
        message += f"""
-- Output captured from stdout:
{e.stdout}"""

    if include_stderr and e.stderr is not None:
        message += f"""
-- Output captured from stderr:
{e.stderr}"""

    return message


def _quote_whitespace(string: str) -> str:
    if " " in string:
        return f"'{string}'"
    else:
        return string



from __future__ import annotations

import subprocess
from setuptools.errors import PlatformError
from functools import lru_cache
from typing import Dict, List, NewType, Optional, TYPE_CHECKING

from ._utils import Env, check_subprocess_output

if TYPE_CHECKING:
    from semantic_version import Version


def get_rust_version(env: Optional[Env]) -> Optional[Version]:  # type: ignore[no-any-unimported]
    try:
        # first line of rustc -Vv is something like
        # rustc 1.61.0 (fe5b13d68 2022-05-18)
        from semantic_version import Version

        return Version(_rust_version(env).split(" ")[1])
    except (subprocess.CalledProcessError, OSError):
        return None


_HOST_LINE_START = "host: "


def get_rust_host(env: Optional[Env]) -> str:
    # rustc -Vv has a line denoting the host which cargo uses to decide the
    # default target, e.g.
    # host: aarch64-apple-darwin
    for line in _rust_version_verbose(env).splitlines():
        if line.startswith(_HOST_LINE_START):
            return line[len(_HOST_LINE_START) :].strip()
    raise PlatformError("Could not determine rust host")


RustCfgs = NewType("RustCfgs", Dict[str, Optional[str]])


def get_rustc_cfgs(target_triple: Optional[str], env: Env) -> RustCfgs:
    cfgs = RustCfgs({})
    for entry in get_rust_target_info(target_triple, env):
        maybe_split = entry.split("=", maxsplit=1)
        if len(maybe_split) == 2:
            cfgs[maybe_split[0]] = maybe_split[1].strip('"')
        else:
            assert len(maybe_split) == 1
            cfgs[maybe_split[0]] = None
    return cfgs


@lru_cache()
def get_rust_target_info(target_triple: Optional[str], env: Env) -> List[str]:
    cmd = ["rustc", "--print", "cfg"]
    if target_triple:
        cmd.extend(["--target", target_triple])
    output = check_subprocess_output(cmd, env=env, text=True)
    return output.splitlines()


@lru_cache()
def get_rust_target_list(env: Env) -> List[str]:
    output = check_subprocess_output(
        ["rustc", "--print", "target-list"], env=env, text=True
    )
    return output.splitlines()


@lru_cache()
def _rust_version(env: Env) -> str:
    return check_subprocess_output(["rustc", "-V"], env=env, text=True)


@lru_cache()
def _rust_version_verbose(env: Env) -> str:
    return check_subprocess_output(["rustc", "-Vv"], env=env, text=True)
